@@ -3,24 +3,31 @@
 Application class to be used from TestSandbox to keep list of running applications.
 """
 
+# pylint: disable=broad-exception-caught
+# pylint: disable=import-outside-toplevel
+# pylint: disable=import-error
+# pylint: disable=invalid-name
+# ruff: noqa: E501
+
 import shlex
 import configparser
 from time import sleep
 from typing import TypeVar
 from subprocess import Popen, PIPE
-from dogtail.rawinput import (  # pylint: disable=import-error
+from dogtail.rawinput import (
     typeText,
     pressKey,
     keyCombo,
 )
-from qecore.utility import run, run_verbose, overview_action
+from qecore.utility import RunAndLog, overview_action, get_func_params_and_values
 from qecore.logger import Logging
 
 logging_class = Logging()
-log = logging_class.logger
+LOGGING = logging_class.logger
 
-# We do this just to satisfy the typing hint but without the need to import entire class.
-gi_repository_Atspi_Accessible = TypeVar(  # pylint: disable=invalid-name
+# We do this just to satisfy the typing hint
+# but without the need to import entire class.
+gi_repository_Atspi_Accessible = TypeVar(
     "gi_repository_Atspi_Accessible"
 )
 
@@ -55,7 +62,7 @@ class Application:
         :param a11y_app_name: Name of application as it appears in accessibility tree.
 
         :type desktop_file_exists: bool
-        :param desktop_file_exists: States that desktop file of given application exists or not.
+        :param desktop_file_exists: Desktop file of given application exists.
 
         :type desktop_file_path: str
         :param desktop_file_path: Full posix path to .desktop file of application.
@@ -67,18 +74,7 @@ class Application:
         :param app_process_name: Name of application process if not same as component.
         """
 
-        log.debug(
-            " ".join(
-                (
-                    f"__init__(self, component={component}, a11y_app_name={a11y_app_name},",
-                    f"desktop_file_exists={desktop_file_exists},",
-                    f"desktop_file_name={desktop_file_name},",
-                    f"desktop_file_path={desktop_file_path},",
-                    f"app_process_name={app_process_name},",
-                    f"session_type={session_type}, session_desktop={session_desktop})",
-                )
-            )
-        )
+        LOGGING.debug(get_func_params_and_values())
 
         # Define shell as None in the moment on initialization.
         self.shell = None
@@ -92,13 +88,17 @@ class Application:
         self.desktop_file_exists = desktop_file_exists
         self.exit_shortcut = "<Control_L><Q>"
         self.kill = True
-        self.kill_command = None
+        self.kill_command = ""
         self.desktop_file_name = desktop_file_name
         self.desktop_file_path = desktop_file_path
         self.application_list = []
         self.app_process_name = app_process_name if app_process_name else component
         self.icon = None
         self.process = None
+
+        self.name = None
+        self.exec = None
+        self.icon = None
         self.get_desktop_file_data()
 
         # To be removed in the future, no more camelCasing.
@@ -114,6 +114,9 @@ class Application:
                 self.get_all_kill_candidates
             )
 
+        self.blocked_list = ["TMT_", "runtest", "behave", "qecore", "dogtail", "gnome-shell", "harness"]
+        self.permitted_list = ["org.gnome.Shell.Extensions", "org.gnome.Extensions"]
+
     def get_desktop_file_data(self) -> None:
         """
         Parse desktop file.
@@ -123,7 +126,7 @@ class Application:
             Do **NOT** call this by yourself. This method is called by :func:`__init__`.
         """
 
-        log.debug(f"{self.component} get_desktop_file_data(self)")
+        LOGGING.debug(get_func_params_and_values() + " for " + self.component)
 
         desktop_file_help = """
         Specify one when initializing your application in environment.py, options are:
@@ -133,69 +136,69 @@ class Application:
         """
 
         # zenity/gnome-shell do not have desktop file.
-        if self.desktop_file_exists:
-            if self.desktop_file_path:
-                log.debug(
-                    " ".join(
-                        (
-                            "get_desktop_file_data: loading user provided .desktop file:",
-                            f"'{self.desktop_file_path}'",
-                        )
-                    )
-                )
-                desktop_file = self.desktop_file_path
-            else:
-                desktop_run = run_verbose(
-                    " ".join(
-                        (
-                            f"rpm -qlf $(which {self.component}) |",
-                            "grep /usr/share/applications/ |",
-                            "grep .desktop |",
-                            f"grep '{self.desktop_file_name}'",
-                        )
-                    )
-                )
+        if not self.desktop_file_exists:
+            LOGGING.debug(f"Not handling desktop file for component '{self.component}'")
+            return
 
-                if desktop_run[1] != 0:
-                    raise UserWarning(
-                        f"Desktop file of application '{self.component}' was not found."
-                    )
-
-                desktop_file = desktop_run[0].strip("\n")
-                log.debug(
-                    f"get_desktop_file_data: .desktop file found: '{desktop_file}'"
-                )
-
-            desktop_files_list = desktop_file.split("\n")
-            if len(desktop_files_list) != 1:
-                assert False, "\n".join(
-                    (
-                        f"More than one .desktop file found: \n{desktop_file}\n",
-                        desktop_file_help,
-                    )
-                )
-
-            desktop_file_config = configparser.RawConfigParser()
-            list_of_successfully_parsed_files = desktop_file_config.read(desktop_file)
-
-            if not list_of_successfully_parsed_files:
-                assert False, "".join(
-                    (f"Failed attepmt to parse the .desktop file: '{desktop_file}'.")
-                )
-
-            self.name = desktop_file_config.get("Desktop Entry", "name")
-            self.exec = desktop_file_config.get("Desktop Entry", "exec").split(" ")[0]
-            self.icon = desktop_file_config.get("Desktop Entry", "icon", fallback=None)
-            log.debug(
+        if self.desktop_file_path:
+            LOGGING.debug(
                 " ".join(
                     (
-                        "get_desktop_file_data:",
-                        f"data name='{self.name}'",
-                        f"exec='{self.exec}'",
-                        f"icon='{self.icon}'",
+                        "loading user provided .desktop file:",
+                        f"'{self.desktop_file_path}'",
                     )
                 )
             )
+            desktop_file = self.desktop_file_path
+        else:
+            desktop_run = RunAndLog(
+                " ".join(
+                    (
+                        f"rpm -qlf $(which {self.component}) |",
+                        "grep /usr/share/applications/ |",
+                        "grep .desktop |",
+                        f"grep '{self.desktop_file_name}'",
+                    )
+                )
+            )
+
+            if desktop_run.return_code != 0:
+                raise UserWarning(
+                    f"Desktop file of application '{self.component}' was not found."
+                )
+
+            desktop_file = desktop_run.output.strip("\n")
+            LOGGING.debug(f".desktop file found: '{desktop_file}'")
+
+        desktop_files_list = desktop_file.split("\n")
+        if len(desktop_files_list) != 1:
+            assert False, "\n".join(
+                (
+                    f"More than one .desktop file found: \n{desktop_file}\n",
+                    desktop_file_help,
+                )
+            )
+
+        desktop_file_config = configparser.RawConfigParser()
+        list_of_successfully_parsed_files = desktop_file_config.read(desktop_file)
+
+        if not list_of_successfully_parsed_files:
+            assert False, "".join(
+                (f"Failed attempt to parse the .desktop file: '{desktop_file}'.")
+            )
+
+        self.name = desktop_file_config.get("Desktop Entry", "name")
+        self.exec = desktop_file_config.get("Desktop Entry", "exec").split(" ")[0]
+        self.icon = desktop_file_config.get("Desktop Entry", "icon", fallback=None)
+        LOGGING.debug(
+            " ".join(
+                (
+                    f"data name='{self.name}'",
+                    f"exec='{self.exec}'",
+                    f"icon='{self.icon}'",
+                )
+            )
+        )
 
     def start_via_command(self, command=None, **kwargs) -> None:
         """
@@ -208,6 +211,8 @@ class Application:
         :param in_session: Start application via command in session.
         """
 
+        LOGGING.debug(get_func_params_and_values() + " for " + self.component)
+
         kill_override = self.kill
         in_session = False
 
@@ -218,18 +223,25 @@ class Application:
             if "kill" in str(key).lower():
                 kill_override = val
 
-        log.debug(
-            "".join(
-                (
-                    f"{self.component} ",
-                    f"start_via_command(self, command={command}, in_session={in_session})",
-                )
-            )
-        )
-
         if self.is_running() and kill_override:
             self.kill_application()
 
+        command_or_exec_help = "\n".join((
+            "You attempted to use 'start_via_command' without any command provided.",
+            "This function gets command from application's .desktop file or from user.",
+            "",
+            "You can either:",
+            "1) In case of the command not changing, define it in your environment.py:",
+            "    context.<application>.exec = \"<command_to_run>\"",
+            "2) In case of the command changing in every test case:",
+            "   * Start application \"<application>\" via command \"<command>\"",
+        ))
+
+        # In a case of an application with no desktop file.
+        # Notify user that they are trying to use incorrect step.
+        if command is None and self.exec is None:
+            raise RuntimeError(command_or_exec_help)
+
         command_to_use = command if command else self.exec
 
         if in_session:
@@ -283,7 +295,7 @@ class Application:
         in a new gnome-classic as it lost the search bar.
         """
 
-        log.debug(f"{self.component} start_via_menu(self)")
+        LOGGING.debug(get_func_params_and_values() + " for " + self.component)
 
         if self.kiosk:
             raise RuntimeError("Start via menu is not available in kiosk mode.")
@@ -295,9 +307,9 @@ class Application:
 
         if not self.desktop_file_exists:
             raise UserWarning(
-                "".join(
+                " ".join(
                     (
-                        f"Testing target '{self.a11y_app_name}' does not have desktop file. ",
+                        f"The target '{self.a11y_app_name}' doesn't have desktop file.",
                         "Indication of user failure!",
                     )
                 )
@@ -322,16 +334,16 @@ class Application:
         Close application via shortcut.
 
         Closing application via shortcut Ctrl+Q.
-        If for any reason the application does not have this shortcut you can define the shortcut
-        to use in class attribute :py:attr:`exit_shortcut`
+        If for any reason the application does not have this shortcut you can
+        define the shortcut to use in class attribute :py:attr:`exit_shortcut`
 
         .. note::
 
-            Format of the shortcut is a string with each key character encapsuled in < >.
+            Format of the shortcut is a string with each key character encapsuled in <>.
             Here are  a few examples: <Ctrl><Q> <Alt><F4> <Ctrl><Shift><Q>
         """
 
-        log.debug(f"{self.component} close_via_shortcut(self)")
+        LOGGING.debug(get_func_params_and_values() + " for " + self.component)
 
         if not self.is_running():
             raise RuntimeWarning(
@@ -354,27 +366,32 @@ class Application:
         this will retrieve the root of the application.
         """
 
-        log.debug(f"{self.component} already_running(self)")
+        LOGGING.debug(get_func_params_and_values() + " for " + self.component)
 
         self.wait_before_app_starts(15)
         self.instance = self.get_root()
 
     def get_root(self) -> gi_repository_Atspi_Accessible:
         """
-        Get accessibility root of appllication.
+        Get accessibility root of application.
 
         :return: Return root object of application.
         :rtype: <dogtail.tree.root.application>
 
-        This method is used upon application start to retrieve a11y object to use in the test.
-        You do not need to use this by yourself, but can be useful in some situations.
+        This method is used upon application start to retrieve a11y object to
+        use in the test. You do not need to use this by yourself, but can be useful
+        in some situations.
         """
 
-        log.debug(f"{self.component} get_root(self) search by '{self.a11y_app_name}'")
+        LOGGING.debug(get_func_params_and_values() + " for " + self.component)
 
-        # Importing here for issue with accessibility if the connection is not ready yet.
+        if self.component != self.a11y_app_name:
+            LOGGING.debug(f"Get root for '{self.component}' with '{self.a11y_app_name}'")
+
+        # Importing here as there is sometimes an issue with accessibility
+        # if the connection is not ready yet.
         # This will make sure the import will happen after the session is loaded.
-        from dogtail.tree import (  # pylint: disable=import-error,import-outside-toplevel
+        from dogtail.tree import (
             root,
         )
 
@@ -383,52 +400,58 @@ class Application:
     def is_running(self) -> bool:
         """
         Get running state of application.
-        This state is retrieved by checking if application is present in the accessibility tree.
+        This state is retrieved by checking if application is present
+        in the accessibility tree.
 
         :return: Return state of application. Running or not.
         :rtype: bool
 
-        This method is used by various other methods in this class to ensure correct behaviour.
-        You do not need to use this by yourself, but can be useful in some situations.
+        This method is used by various other methods in this class to ensure correct
+        behavior. You do not need to use this by yourself, but can be useful in some
+        situations.
         """
 
-        log.debug(f"{self.component} is_running(self)")
+        LOGGING.debug(get_func_params_and_values() + " for " + self.component)
 
         is_running = False
 
-        try:
-            # Preventing issues of loading a11y on import before session is ready.
-            from dogtail.tree import (  # pylint: disable=import-error,import-outside-toplevel
-                root,
-            )
+        # Give an application list a few more tries to load. Following error exists:
+        # GLib.GError: atspi_error: The application no longer exists (0)
+        for _ in range(3):
+            try:
+                # Preventing issues of loading a11y on import before session is ready.
+                from dogtail.tree import (
+                    root,
+                )
 
-            self.application_list = [x.name for x in root.applications()]
-            for _ in range(3):
+                self.application_list = [x.name for x in root.applications()]
                 is_running = self.a11y_app_name in self.application_list
-                if is_running:
-                    break
-            return is_running
-        except RuntimeError:
-            return False
+                LOGGING.debug(f"{self.component} Atspi tree check: '{is_running}'")
+                return is_running
+
+            except RuntimeError:
+                LOGGING.debug(f"{self.component} is_running(self) - RuntimeError")
+                return False
 
     def get_pid_list(self) -> str:
         """
         Get list of processes of running application.
 
         :return: Return all running processes of application, \
-            seperated by new line character, not converting to list. \
+            separated by new line character, not converting to list. \
             Return nothing if application is not running.
 
         .. note::
 
             Do **NOT** call this by yourself.
-            This method is called by :func:`get_all_kill_candidates` when stopping the application.
+            This method is called by :func:`get_all_kill_candidates`
+            when stopping the application.
         """
 
-        log.debug(f"{self.component} get_pid_list(self)")
+        LOGGING.debug(get_func_params_and_values() + " for " + self.component)
 
         if self.is_running():
-            return run(f"pgrep -fla {self.app_process_name}").strip("\n")
+            return RunAndLog(f"pgrep -fla {self.app_process_name}").output.strip("\n")
 
         return ""
 
@@ -442,15 +465,18 @@ class Application:
         .. note::
 
             Do **NOT** call this by yourself.
-            This method is called by :func:`kill_application` when stopping the application.
+            This method is called by :func:`kill_application`
+            when stopping the application.
 
-        If kill candidate happens to have ['runtest', 'behave', 'dogtail', '/usr/bin/gnome-shell']
+        If kill candidate happens to have
+        ['runtest', 'behave', 'dogtail', '/usr/bin/gnome-shell']
         in its process name. Process will not be killed.
         Return empty list if no satisfactory process was found.
-        This prevents testname colliding with found process so we will not kill the test itself.
+        This prevents test name colliding with found process so that
+        we will not kill the test itself.
         """
 
-        log.debug(f"{self.component} get_all_kill_candidates(self)")
+        LOGGING.debug(get_func_params_and_values() + " for " + self.component)
 
         application_pid_string = self.get_pid_list()
         if application_pid_string:
@@ -460,39 +486,52 @@ class Application:
 
         final_pid_list = []
         for process in application_pid_list:
-            if not (
-                ("runtest" in process)
-                or ("behave" in process)
-                or ("dogtail" in process)
-                or (process == "/usr/bin/gnome-shell")
-            ):
+            LOGGING.debug(f"Evaluation of process: '{process}'")
+
+            # Evaluate the block and allow lists.
+            block_the_cleanup = [blocked_item in process for blocked_item in self.blocked_list]
+            allow_the_cleanup = [allowed_item in process for allowed_item in self.permitted_list]
+
+            # Block list evaluation for logging to quickly debug things.
+            LOGGING.debug(f"Block list: '{self.blocked_list}'")
+            LOGGING.debug(f"Block evaluation: '{block_the_cleanup}'")
+            # Allow list evaluation for logging to quickly debug things.
+            LOGGING.debug(f"Allow list: '{self.permitted_list}'")
+            LOGGING.debug(f"Allow evaluation: '{allow_the_cleanup}'")
+
+            # We cleanup only on 'no block' or with 'allow' even if block was made.
+            if not any(block_the_cleanup) or any(allow_the_cleanup):
+                LOGGING.debug(f"Adding '{process}' to cleanup list.")
                 extracted_pid = process.split(" ", 1)[0]
                 try:
                     final_pid_list.append(int(extracted_pid))
                 except ValueError:
                     # Skip non-digits.
                     pass
+            else:
+                LOGGING.debug(f"Attempt to add '{process}' to cleanup list was blocked.")
 
+        LOGGING.debug(f"Final pid list: '{final_pid_list}'")
         return final_pid_list
 
     def kill_application(self) -> None:
         """
         Kill application.
 
-        This method is used by :func:`start_via_command` and :func:`start_via_menu` to ensure
-        the application is not running when starting the test. So we kill all application instances
-        and open a new one to test on.
+        This method is used by :func:`start_via_command` and :func:`start_via_menu`
+        to ensure the application is not running when starting the test.
+        So we kill all application instances and open a new one to test on.
         You do not need to use this by yourself, but can be useful in some situations.
         """
 
-        log.debug(f"{self.component} kill(self)")
+        LOGGING.debug(get_func_params_and_values() + " for " + self.component)
 
         if self.is_running() and self.kill:
             if not self.kill_command:
                 for pid in self.get_all_kill_candidates():
-                    run(f"sudo kill -9 {pid} > /dev/null")
+                    RunAndLog(f"sudo kill -9 {pid} > /dev/null")
             else:
-                run(self.kill_command)
+                RunAndLog(self.kill_command)
 
     # Following two could be merged, I could not think of a nice way of doing it though.
     def wait_before_app_starts(self, time_limit) -> None:
@@ -503,14 +542,7 @@ class Application:
         :param time_limit: Number which signifies time before the run is stopped.
         """
 
-        log.debug(
-            "".join(
-                (
-                    f"{self.component} ",
-                    f"wait_before_app_starts(self, time_limit={time_limit})",
-                )
-            )
-        )
+        LOGGING.debug(get_func_params_and_values() + " for " + self.component)
 
         for _ in range(time_limit * 10):
             if not self.is_running():
@@ -533,14 +565,7 @@ class Application:
         :param time_limit: Number which signifies time before the run is stopped.
         """
 
-        log.debug(
-            "".join(
-                (
-                    f"{self.component} ",
-                    f"wait_before_app_closes(self, time_limit={time_limit}",
-                )
-            )
-        )
+        LOGGING.debug(get_func_params_and_values() + " for " + self.component)
 
         for _ in range(time_limit * 10):
             if self.is_running():
