@@ -3,11 +3,21 @@
 Get Node class to be used for common_steps.
 """
 
+# pylint: disable=broad-exception-caught
+# pylint: disable=import-outside-toplevel
+# pylint: disable=import-error
+# ruff: noqa: E501
+
 import sys
 import traceback
 import re
 
-from qecore.utility import QE_DEVELOPMENT, Tuple
+from qecore.logger import Logging
+from qecore.utility import Tuple
+
+logging_class = Logging()
+LOGGING = logging_class.logger
+
 
 # For python 3.6.
 if not hasattr(re, "Pattern"):
@@ -18,12 +28,6 @@ Filip Pokryvka <fpokryvk@redhat.com>,
 Michal Odehnal <modehnal@redhat.com>
 """
 
-if QE_DEVELOPMENT:
-    try:
-        from termcolor import colored
-    except ModuleNotFoundError:
-        QE_DEVELOPMENT = False
-
 
 class GetNode:
     """
@@ -47,6 +51,12 @@ class GetNode:
         :type description: str
         :param description: Node.description
 
+        :type action: str
+        :param action: Node.actions
+
+        :type offset: str
+        :param offset: Offset to use before click is executed.
+
         :type m_btn: str
         :param m_btn: Mouse click after node identification.
             Accepted values are "Left", "Middle" and "Right".
@@ -57,19 +67,19 @@ class GetNode:
 
         :type size_low: string
         :param size_low: minimum node size
-            Should be formated "width,height", where width and height are integers
+            Should be formatted "width,height", where width and height are integers
 
         :type size_high: string
         :param size_high: maximum node size
-            Should be formated "width,height", where width and height are integers
+            Should be formatted "width,height", where width and height are integers
 
         :type position_low: string
-        :param position_low: minimum distance from topleft corner
-            Should be formated "x,y", where x and y are integers
+        :param position_low: minimum distance from top left corner
+            Should be formatted "x,y", where x and y are integers
 
         :type position_high: string
-        :param position_high: maximum distance from topleft corner
-            Should be formated "x,y", where x and y are integers
+        :param position_high: maximum distance from top left corner
+            Should be formatted "x,y", where x and y are integers
 
         :type a11y_root_name: str
         :param a11y_root_name: Application name.
@@ -80,7 +90,7 @@ class GetNode:
         :type retry: bool
         :param retry: Option to give search function to look again a few times if search fails.
             Used for slower applications. User might want to click right away but window can
-            have a few secods delay.
+            have a few seconds delay.
 
         :type expect_positive: bool
         :param expect_positive: Option to pass the common step call if the node is not found.
@@ -92,13 +102,19 @@ class GetNode:
         """
 
         # Importing here to prevent earlier import error when session is not ready yet.
-        from dogtail.tree import root  # pylint: disable=import-error
+        from dogtail.tree import root
 
         name = kwargs.get("name", None)
         role_name = kwargs.get("role_name", None)
         m_btn = kwargs.get("m_btn", None)
         attr = kwargs.get("attr", None)
         description = kwargs.get("description", None)
+        action = kwargs.get("action", None)
+
+        offset = kwargs.get("offset", "0,0")
+
+        self.x_node_center = 0
+        self.y_node_center = 0
 
         def string_to_int_tuple(arg) -> Tuple:
             return Tuple((int(x) for x in arg.split(",")))
@@ -118,7 +134,16 @@ class GetNode:
         if a11y_root_name is not None:
             # Try to find root application name in context.
             if hasattr(context, a11y_root_name):
-                self.root = getattr(context, a11y_root_name).instance
+                # If the user runs action on application lets assume it is running.
+                # Save the application class instance.
+                application_class_instance = getattr(context, a11y_root_name)
+                # The a11y instance does not have to be loaded right away, try to load it.
+                if not application_class_instance.instance:
+                    # This is needed to get a11y root if
+                    # application was not started by common steps.
+                    application_class_instance.already_running()
+                # Instance should be loaded so we can load the root.
+                self.root = application_class_instance.instance
             # Try to find root application name in accessibility tree.
             else:
                 try:
@@ -137,7 +162,15 @@ class GetNode:
         else:
             # Only option for not having the name is to have default application set.
             try:
-                self.root = context.sandbox.default_application.instance
+                # If the user runs action on application lets assume it is running.
+                default_application_class_instance = context.sandbox.default_application
+                # The a11y instance does not have to be loaded right away, try to load it.
+                if not default_application_class_instance.instance:
+                    # This is needed to get a11y root if
+                    # application was not started by common steps.
+                    default_application_class_instance.already_running()
+                # Instance should be loaded so we can load the root.
+                self.root = default_application_class_instance.instance
             except AttributeError:
                 traceback.print_exc(file=sys.stdout)
                 assert False, "".join(
@@ -147,8 +180,46 @@ class GetNode:
                     )
                 )
 
-        # Condition to do any action on any node, is that the application is running.
-        application_list = [x.name for x in root.applications()]
+        # Attempt to prevent the following error when loading the application.
+        # GLib.GError: atspi_error: The application no longer exists (0)
+        # Variable for all application names.
+        application_list = []
+        # Variable for all accessibility objects for name loading later.
+        application_objects = []
+        # Set a flag to not spam console in positive cases.
+        exception_occurred = False
+        for _ in range(5):
+            try:
+                # Load all application objects.
+                application_objects = root.applications()
+            except Exception as error:
+                # Log to the console right away proving the error comes from this part.
+                LOGGING.debug(f"Exception thrown when loading all applications: '{error}'")
+            else:
+                # Execute only when the exception is not thrown.
+                # Break the cycle if the application objects are loaded.
+                break
+
+        # Iterate over all application objects and load application names to the list.
+        for _ in range(5):
+            for application in application_objects:
+                try:
+                    # Load the name of the object to the list.
+                    application_list.append(application.name)
+                except Exception as error:
+                    exception_occurred = True
+                    LOGGING.debug(f"Exception thrown while getting object name: '{error}'")
+                    # Append dummy string to show something went wrong.
+                    application_list.append("<exception_occurrence>")
+
+            # Break the cycle if the application names are loaded.
+            break
+
+        # Log the final application list only when error occurred when loading names.
+        if exception_occurred:
+            LOGGING.debug(f"Final application list: '{application_list}'")
+
+        # Check if the root of our application is in the list.
         if self.root.name not in application_list:
             assert False, "".join(
                 (
@@ -158,6 +229,7 @@ class GetNode:
                 )
             )
 
+        # Validate various options.
         mouse_map = {"Left": 1, "Middle": 2, "Right": 3, "None": None}
         try:
             self.m_btn = mouse_map[str(m_btn)]
@@ -177,6 +249,21 @@ class GetNode:
         self.name = string_to_regex(name, flags=re.DOTALL | re.MULTILINE)
         self.role_name = string_to_regex(role_name)
         self.description = string_to_regex(description, flags=re.DOTALL | re.MULTILINE)
+        self.action = string_to_regex(action, flags=re.DOTALL | re.MULTILINE)
+
+        try:
+            # Expected format is "(X,Y)" but for example "X, Y" will work too.
+            self.offset = offset.lstrip("(").rstrip(")").replace(" ", "").split(",")
+            self.offset_x = int(self.offset[0])
+            self.offset_y = int(self.offset[1])
+
+        except Exception as error:
+            self.offset = [0, 0]
+            self.offset_x = 0
+            self.offset_y = 0
+
+            LOGGING.debug(f"Offset provided in unexpected format: '{error}'")
+            LOGGING.debug("Expected format is '(X,Y)'.")
 
         defined_attributes = [
             "showing",
@@ -236,12 +323,16 @@ class GetNode:
                 )
                 and (
                     (not isinstance(self.description, str))
-                    or self.description == x.description
+                    or self.description in repr(x.description)
                 )
                 and (
                     (not isinstance(self.description, re.Pattern))
                     or re.fullmatch(self.description, x.description)
                 )
+                and (
+                    (not isinstance(self.action, str))
+                    or self.action in x.actions
+                )
                 and ((self.attr is None) or getattr(x, self.attr))
                 and Tuple(x.position) >= self.position_low
                 and Tuple(x.position) <= self.position_high
@@ -249,39 +340,26 @@ class GetNode:
                 and Tuple(x.size) <= self.size_high,
                 retry=self.retry,
             )
-        except Exception as error:  # pylint: disable=broad-except
+        except Exception as error:
             if self.expect_positive:
                 assert False, self.get_error_message(self, error)
             else:
-                found_node = None
+                # When not expecting positives, lets return the result right away.
+                return (self, None)
+
+        # Once the Node is found lets get the center with offset for better manipulation later.
+        self.x_node_center = int(found_node.position[0] + found_node.size[0] / 2 + self.offset_x)
+        self.y_node_center = int(found_node.position[1] + found_node.size[1] / 2 + self.offset_y)
+
         return (self, found_node)
 
-    def __exit__(self, exc_type, exc_value, trcb) -> bool:
+    def __exit__(self, exc_type, exc_value, trace_b) -> bool:
         if exc_type is not None:
             return False
         return True
 
-    def format_string(self, string, bold=False) -> str:
-        """
-        Colour string based on environemnt.
 
-        :param string: String to format.
-        :type string: str
-
-        :param bold: Bold text.
-        :type bold: bool
-
-        :return: Formatted string.
-        :rtype: str
-        """
-
-        if bold:
-            return (
-                colored(string, "yellow", attrs=["bold"]) if QE_DEVELOPMENT else string
-            )
-        return colored(string, "yellow") if QE_DEVELOPMENT else string
-
-    def get_formated_duplicates(self, list_size, list_of_duplicates, attr) -> str:
+    def get_formatted_duplicates(self, list_size, list_of_duplicates, attr) -> str:
         """
         Take list of duplicates and format them for error message.
 
@@ -303,18 +381,20 @@ class GetNode:
             Used by :py:func:`get_error_message`.
         """
 
+        tab_space = " " * 4
+
         return (
             "".join(
                 sorted(
                     {
                         "".join(
                             (
-                                "\t",
-                                f"{self.format_string('name')}: '{repr(node.name)}' ",
-                                f"{self.format_string('roleName')}: '{node.roleName}' ",
-                                f"{self.format_string('position')}: '{node.position}' ",
-                                f"{self.format_string('size')}: '{node.size}' ",
-                                f"{self.format_string(attr if attr else 'attribute')}: ",
+                                tab_space,
+                                f"{'name'}: {repr(node.name)} ",
+                                f"{'roleName'}: '{node.roleName}' ",
+                                f"{'position'}: '{node.position}' ",
+                                f"{'size'}: '{node.size}' ",
+                                f"{attr if attr else 'attribute'}: ",
                                 f"'{getattr(node, attr) if attr else 'None'}'",
                                 "\n",
                             )
@@ -325,10 +405,10 @@ class GetNode:
                 )
             )
             if list_size != 0
-            else "\tNone\n"
+            else tab_space + "None\n"
         )
 
-    def get_formated_error_message(
+    def get_formatted_error_message(
         self, error, node_name, same_name_items, node_role_name, same_role_name_items
     ) -> str:
         """
@@ -350,7 +430,7 @@ class GetNode:
         :param same_role_name_items: List of all items with the roleName node_role_name.
 
         :rtype: string
-        :return: Formatted string output of all :py:func:`get_formated_duplicates`
+        :return: Formatted string output of all :py:func:`get_formatted_duplicates`
 
         .. note::
 
@@ -361,13 +441,13 @@ class GetNode:
         return "".join(
             (
                 "\n\n",
-                f"{self.format_string('Item was not found', bold=True)}: {error}",
+                f"Item was not found: {error}",
                 "\n\n",
-                f"{self.format_string('Items with name', bold=True)}: {node_name}:",
+                f"Items with name: {node_name}:",
                 "\n",
                 same_name_items,
                 "\n",
-                f"{self.format_string('Items with roleName', bold=True)}: {node_role_name}:",
+                f"Items with roleName: {node_role_name}:",
                 "\n",
                 same_role_name_items,
                 "\n",
@@ -376,8 +456,8 @@ class GetNode:
 
     def get_error_message(self, node, error) -> str:
         """
-        Main handler for error message with :py:func:`get_formated_error_message` and
-        :py:func:`get_formated_duplicates` being used to get desired result.
+        Main handler for error message with :py:func:`get_formatted_error_message` and
+        :py:func:`get_formatted_duplicates` being used to get desired result.
 
         :type node: GetNode
         :param node: Instanced GetNode to have all data needed about the error.
@@ -404,7 +484,7 @@ class GetNode:
                 lambda x: re.fullmatch(node.name, x.name)
             )
         nodes_with_name_size = len(nodes_with_name)
-        nodes_with_name_formatted = self.get_formated_duplicates(
+        nodes_with_name_formatted = self.get_formatted_duplicates(
             nodes_with_name_size, nodes_with_name, node.attr
         )
         role_name = repr(node.role_name)
@@ -419,11 +499,11 @@ class GetNode:
                 lambda x: re.fullmatch(node.role_name, x.roleName)
             )
         nodes_with_role_name_size = len(nodes_with_role_name)
-        nodes_with_role_name_formatted = self.get_formated_duplicates(
+        nodes_with_role_name_formatted = self.get_formatted_duplicates(
             nodes_with_role_name_size, nodes_with_role_name, node.attr
         )
 
-        return self.get_formated_error_message(
+        return self.get_formatted_error_message(
             error,
             name,
             nodes_with_name_formatted,
